import operator
from collections import defaultdict
from functools import partialmethod, total_ordering
from typing import Iterable, List, Optional, Union

from mido import Message
from sortedcontainers import SortedSet

from ..constants import DEFAULT_TICKS_PER_BEAT, KEYS
from ..warps import optional_inplace
from .event import NoteOff, NoteOn

__all__ = [
    "Note",
    "NoteSet",
]


@total_ordering
class Note:

    def __init__(
        self,
        note: Union[str, int, float],
        velocity: int = 127,
        length: float = 1.,
        location: float = 0.,
        channel: int = 0,
    ) -> None:
        self._note_on = NoteOn(note,
                               velocity=velocity,
                               location=location,
                               channel=channel)
        self.length = length

    @property
    def name(self) -> str:
        return self._note_on.name

    @name.setter
    def name(self, other: str) -> None:
        self._note_on.name = other

    @property
    def value(self) -> int:
        return self._note_on.value

    @value.setter
    def value(self, other: int) -> None:
        self._note_on.value = other

    @property
    def frequency(self) -> float:
        return self._note_on.frequency

    @frequency.setter
    def frequency(self, other: float) -> None:
        self.frequency = other

    @property
    def key(self) -> float:
        return self._note_on.key

    @property
    def octave(self) -> float:
        return self._note_on.octave

    @property
    def velocity(self) -> int:
        return self._note_on.velocity

    @velocity.setter
    def velocity(self, value: int) -> None:
        self._note_on.velocity = value

    @property
    def location(self) -> float:
        return self._note_on.location

    @location.setter
    def location(self, value: float) -> None:
        self._note_on.location = value

    @property
    def channel(self) -> int:
        return self._note_on.channel

    @channel.setter
    def channel(self, channel: int) -> None:
        self._note_on.channel = channel

    def copy(self) -> 'Note':
        return self.__class__(self.value,
                              velocity=self.velocity,
                              length=self.length,
                              location=self.location,
                              channel=self.channel)

    def _compare(self, a, b, op):
        _a = (a.location, a.value, a.channel)
        _b = (b.location, b.value, b.channel)
        return op(_a, _b)

    def __lt__(self, other: 'Note') -> bool:
        return self._compare(self, other, operator.lt)

    def __eq__(self, other: 'Note') -> bool:
        return self._compare(self, other, operator.eq)

    def __hash__(self):
        return hash((self.__class__, self.value, self.location, self.channel))

    def shift(self, length: float = 0) -> 'Note':
        self.location += length
        return self

    @optional_inplace(True)
    def lshift(self, length: float = 0) -> 'Note':
        return self.shift(-length)

    __lshift__ = partialmethod(lshift, inplace=False)
    __ilshift__ = partialmethod(lshift, inplace=True)

    @optional_inplace(True)
    def rshift(self, length: float = 0) -> 'Note':
        return self.shift(length)

    __rshift__ = partialmethod(rshift, inplace=False)
    __irshift__ = partialmethod(rshift, inplace=True)

    @optional_inplace(True)
    def up(self, n: int) -> 'Note':
        self._note_on.value += n
        return self

    __add__ = partialmethod(up, inplace=False)
    __iadd__ = partialmethod(up, inplace=True)

    @optional_inplace(True)
    def down(self, n: int) -> 'Note':
        self._note_on.value -= n
        return self

    __sub__ = partialmethod(down, inplace=False)
    __isub__ = partialmethod(down, inplace=True)

    @optional_inplace(True)
    def zoom(self, ration: float) -> 'Note':
        self.location *= ration
        self.length *= ration
        return self

    __xor__ = partialmethod(zoom, inplace=False)
    __ixor__ = partialmethod(zoom, inplace=True)

    @optional_inplace(True)
    def scale(self, ration: float) -> 'Note':
        self.length *= ration
        return self

    __mul__ = partialmethod(scale, inplace=False)
    __imul__ = partialmethod(scale, inplace=True)

    @optional_inplace(True)
    def decay(self, ration: float) -> 'Note':
        self.velocity = int(self.velocity * ration)
        return self

    __and__ = partialmethod(decay, inplace=False)
    __iand__ = partialmethod(decay, inplace=True)

    @optional_inplace(True)
    def at(self, channel: int) -> 'Note':
        self._note_on.channel = channel
        return self

    __matmul__ = partialmethod(at, inplace=False)
    __imatmul__ = partialmethod(at, inplace=True)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'Note(name="{self.name}", value={self.value}, frequency={self.frequency:.2f}, velocity={self.velocity}, length={self.length:.2f}, location={self.location:.2f}, channel={self.channel})'

    @property
    def note_on(self) -> NoteOn:
        return self._note_on

    @property
    def note_off(self) -> NoteOff:
        return NoteOff(self.value,
                       location=self.location + self.length,
                       velocity=0,
                       channel=self.channel)

    @property
    def msg_on(self) -> Message:
        return self._note_on.msg

    @property
    def msg_off(self) -> Message:
        return self.note_off.msg

    @property
    def msgs(self) -> List[Message]:
        return [self.msg_on, self.msg_off]

    def messages(
            self,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> List[Message]:
        msg_on = self.note_on.message(ticks_per_beat)
        msg_off = self.note_off.message(ticks_per_beat)
        return [msg_on, msg_off]


class _FixedNoteSet(SortedSet):

    def _fix_note(self, note: Note) -> None:
        idx = self.index(note)
        if idx > 0:
            last = self[idx - 1]
            last.length = min(last.length, note.location - last.location)
        if idx < len(self) - 1:
            next = self[idx + 1]
            note.length = min(note.length, next.location - note.location)

    def add(self, note: Note) -> None:
        if isinstance(note, Note):
            if note in self:
                self.discard(note)
            super().add(note)
            self._fix_note(note)

    _add = add


class NoteSet(SortedSet):

    def __init__(self, iterable: Optional[Iterable[Note]] = None, key=None):
        self.mat = defaultdict(lambda: defaultdict(_FixedNoteSet))
        super().__init__(iterable=iterable, key=key)

    def __iter__(self):
        for note in self._list:
            yield note
        # the yielded note can be modifield, to update the order of notes in SortedSet
        self.fix()

    # borrowed from mido
    def __repr__(self) -> str:
        if len(self) == 0:
            notes = ''
        elif len(self) == 1:
            notes = f'[{repr(self[0])}]'
        else:
            notes = f',\n  '.join(repr(m) for m in self)
            notes = f'[\n  {notes}\n]'
        return f'{self.__class__.__name__}({notes})'

    __str__ = __repr__

    def fix(self) -> None:
        notes = list(self._list)
        self.clear()
        self.update(notes)

    def update(self, notes: List[Note]) -> None:
        super().update(notes)
        for note in notes:
            self.mat[note.channel][note.value].add(note)
    __ior__ = update
    _update = update

    def add(self, note: Note) -> None:
        if isinstance(note, Note):
            if note in self:
                self.discard(note)
            super().add(note)
            self.mat[note.channel][note.value].add(note)

    _add = add

    def clear(self) -> None:
        super().clear()
        self.mat = defaultdict(lambda: defaultdict(_FixedNoteSet))

    def copy(self) -> None:
        notes = super().copy()
        notes.mat = self.mat.copy()
        return notes

    def pop(self, index: int = -1) -> Note:
        note = super().pop(index)
        self.mat[note.channel][note.value].remove(note)
        return note

    def discard(self, note: Note) -> None:
        super().discard(note)
        self.mat[note.channel][note.value].discard(note)

    def remove(self, note: Note) -> None:
        super().remove(note)
        self.mat[note.channel][note.value].remove(note)



from ..core import Clip
from ..warps import optional_inplace

__all__ = ["delay"]

@optional_inplace(False)
def delay(clip: Clip, n: int = 1, length: float = 0.1, decay: float= 1.) -> Clip:
    for _ in range(n):
        clip += (clip & decay) >> (length / n)
    return clip


from typing import Iterable, List, Optional, Union

from mido import Message, MetaMessage

from ..constants import DEFAULT_TICKS_PER_BEAT
from .clip import Clip
from .event import Event, ProgramChange
from .note import Note

__all__ = ["Track"]


class Track(Clip):

    def __init__(self,
                 items: Optional[Iterable[Union[Note, Event, 'Clip']]] = None,
                 name: str = '',
                 instrument: Optional[Union[int, str]] = 0,
                 is_drum: bool = False,
                 channel: Optional[int] = 0,
                 mute: bool = False,
                 **kwargs) -> None:
        super().__init__(items=items, **kwargs)
        self.name = name
        self._instrument_id = ProgramChange(
            instrument).id if instrument else None
        self.channel = channel
        if is_drum:
            self.channel = 9
        self.mute = mute

    @property
    def clip(self) -> Clip:
        _clip = Clip()
        _clip.notes = self.notes
        _clip.events = self.events
        return _clip

    @property
    def instrument(self) -> str:
        return ProgramChange(
            self._instrument_id).name if self._instrument_id else ""

    @instrument.setter
    def instrument(self, other: Optional[Union[int, str]] = 0) -> None:
        self._instrument_id = ProgramChange(other).id if other else None

    @property
    def is_drum(self):
        return self.channel == 9

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name="{self.name}", instrument="{self.instrument}", n_notes={self.n_notes}, n_events={self.n_events}, length={self.length}, mute={self.mute}, is_drum={self.is_drum}, channel={self.channel})'

    def copy(self) -> 'Track':
        trk = self.__class__(name=self.name,
                             instrument=self.instrument,
                             channel=self.channel,
                             mute=self.mute)
        trk.notes = self.copy_notes()
        trk.events = self.copy_events()
        return trk

    @property
    def msgs(self):
        return self.messages()

    def messages(
            self,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> List[Message]:
        msgs = []
        if self.name:
            msgs.append(MetaMessage('track_name', name=self.name, time=0))

        if not self.mute:
            if (not self.channel is None) and self._instrument_id:
                msgs.append(
                    ProgramChange(
                        self._instrument_id,
                        channel=self.channel).message(ticks_per_beat))
            msgs.extend(super().messages(ticks_per_beat))
            if not self.channel is None:
                for msg in msgs:
                    if 'channel' in msg.__dict__:
                        msg.channel = self.channel

        msgs.append(MetaMessage('end_of_track', time=0))
        return msgs


if __name__ == '__main__':
    Track()



from ast import Slice
from functools import partialmethod
from typing import List, Optional, Sequence, Union

from mido import Message, MetaMessage, MidiFile, MidiTrack, tempo2bpm

from .. import core
from ..constants import DEFAULT_TICKS_PER_BEAT
from ..exceptions import MIDITypeError, ParameterError
from ..warps import optional_inplace
from .event import Event, EventSet, KeySignature, SetBpm, TimeSignature
from .note import Note
from .track import Track

__all__ = ["Midi"]


class Midi:

    def __init__(self,
                 tracks: Optional[List[Track]] = None,
                 bpm: float = 120,
                 name: str = '',
                 time_signature: Optional[str] = None,
                 key_signature: Optional[str] = None,
                 type: int = 1,
                 ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT,
                 **kwargs):
        if tracks:
            self.tracks = tracks
        else:
            self.tracks = []

        self.bpm = bpm
        self.name = name
        self.time_signature = time_signature
        self.key_signature = key_signature
        self.metaevents = EventSet()
        self.type = type
        self.ticks_per_beat = ticks_per_beat

    def __getitem__(self, key: Union[int, Slice]):
        if isinstance(key, int):
            return self.tracks.__getitem__(key)
        elif isinstance(key, slice):
            return self.slice(key.start, key.stop, inplace=False)
        else:
            raise ParameterError(f"Invalid key type : {type(key)}.")

    def __repr__(self):
        if self.tracks:
            K = len(str(len(self.tracks)))
            tracks_str = ',\n'.join(
                repr(track).replace('Track', 'Track' + '0' *
                                    (K - len(str(i))) + str(i))
                for i, track in enumerate(self.tracks))
            tracks_str = '    ' + tracks_str.replace('\n', '\n    ')
            tracks_str = '\n  tracks=[\n{}\n  ]\n'.format(tracks_str)
        else:
            tracks_str = ''

        return f'{self.__class__.__name__}(name="{self.name}", bpm={self.bpm:.2f}, time_signature="{self.time_signature}", key_signature="{self.key_signature}", ticks_per_beat={self.ticks_per_beat}, {tracks_str})'

    @property
    def length(self) -> int:
        return max(track.length for track in self)

    @optional_inplace(True)
    def trans(self, semi: int) -> 'Midi':
        for track in self:
            if not track.is_drum:
                track += semi
        return self

    @optional_inplace(True)
    def add(self, item: Union[int, Event]):
        if isinstance(item, int):
            self.trans(item)
        elif isinstance(item, Event):
            self.metaevents.add(item)
        else:
            raise ParameterError(f"Unsupported item type: {type(item)}")

    __add__ = partialmethod(add, inplace=False)
    __iadd__ = partialmethod(add, inplace=True)



    @optional_inplace(True)
    def mute(self, arr: Sequence[int]):
        for track in self:
            track.mute = False

        for i in arr:
            self[i].mute = True
        return self

    @optional_inplace(True)
    def unmute(self, arr: Sequence[int]):
        for track in self:
            track.mute = True

        for i in arr:
            self[i].mute = False
        return self

    @optional_inplace(True)
    def slice(self,
              lloc: Optional[float] = None,
              rloc: Optional[float] = None) -> 'Midi':
        for track in self:
            track.slice(lloc, rloc)
        return self

    @optional_inplace(True)
    def repeat(self, n: int) -> 'Midi':
        for track in self:
            track **= n
        return self

    __pow__ = partialmethod(repeat, inplace=False)
    __ipow__ = partialmethod(repeat, inplace=True)

    def shift(self, length: float) -> 'Midi':
        for track in self:
            track >>= length
        return self

    @optional_inplace(True)
    def lshift(self, length: float) -> 'Midi':
        self.shift(-length)
        return self

    __lshift__ = partialmethod(lshift, inplace=False)
    __ilshift__ = partialmethod(lshift, inplace=True)

    @optional_inplace(True)
    def rshift(self, length: float) -> 'Midi':
        self.shift(length)
        return self

    __rshift__ = partialmethod(rshift, inplace=False)
    __irshift__ = partialmethod(rshift, inplace=True)

    @optional_inplace(True)
    def reverse(self) -> 'Midi':
        length = self.length
        for track in self:
            for item in track:
                item.location = length - item.location
        return self

    __reversed__ = partialmethod(reverse, inplace=False)

    @optional_inplace(True)
    def invert(self) -> 'Midi':
        for track in self:
            track.invert()
        return self

    __invert__ = partialmethod(invert, inplace=False)

    @optional_inplace(True)
    def zoom(self, ration: float) -> 'Midi':
        for track in self:
            track ^= ration
        return self

    __xor__ = partialmethod(zoom, inplace=False)
    __ixor__ = partialmethod(zoom, inplace=True)

    @optional_inplace(True)
    def scale(self, ration: float) -> 'Midi':
        for track in self:
            track *= ration
        return self

    __mul__ = partialmethod(scale, inplace=False)
    __imul__ = partialmethod(scale, inplace=True)

    @optional_inplace(True)
    def decay(self, ration: float) -> 'Midi':
        for track in self:
            track &= ration
        return self

    __and__ = partialmethod(decay, inplace=False)
    __iand__ = partialmethod(decay, inplace=True)

    def copy(self) -> 'Midi':
        return Midi([trk.copy() for trk in self.tracks],
                    bpm=self.bpm,
                    time_signature=self.time_signature,
                    key_signature=self.key_signature,
                    ticks_per_beat=self.ticks_per_beat)

    @property
    def numerator(self) -> Optional[int]:
        if self.time_signature:
            return int(self.time_signature.split('/')[0])
        else:
            return None

    @property
    def denominator(self) -> Optional[int]:
        if self.time_signature:
            return int(self.time_signature.split('/')[1])
        else:
            return None

    def to_mido(self) -> MidiFile:
        mid = MidiFile(type=1, ticks_per_beat=self.ticks_per_beat)

        # generate first midi track
        mido_track = MidiTrack()
        mido_track.append(SetBpm(self.bpm).message(self.ticks_per_beat))
        if self.time_signature:
            mido_track.append(
                TimeSignature(self.time_signature).message(
                    self.ticks_per_beat))
        if self.key_signature:
            mido_track.append(
                KeySignature(self.key_signature).message(self.ticks_per_beat))
        if self.name:
            mido_track.append(MetaMessage('track_name', name=self.name,
                                          time=0))

        cur_loc = 0
        for event in self.metaevents:
            event.location, cur_loc = event.location - cur_loc, event.location
            mido_track.append(event.message(self.ticks_per_beat))

        mido_track.append(MetaMessage('end_of_track', time=0))
        mid.tracks.append(mido_track)

        for track in self.tracks:
            mido_track = MidiTrack()
            mido_track.extend(
                track.messages(ticks_per_beat=self.ticks_per_beat))
            mid.tracks.append(mido_track)
        return mid

    @classmethod
    def from_mido(cls, midi_file: MidiFile) -> 'Midi':

        def handel_message(msg, note_msgs: List[List[Optional[Message]]],
                           mid: 'Midi', track: Track, is_first_track: bool):
            event = core.convert.message2event(msg, mid.ticks_per_beat)

            if not event:
                return

            if msg.type in set(('note_off', 'note_on')):
                if msg.type == 'note_off' or (msg.type == 'note_on'
                                              and msg.velocity == 0):
                    if note_msgs[msg.channel][msg.note]:
                        note_on = note_msgs[msg.channel][msg.note]
                        location = core.convert.tick2loc(
                            note_on.time, mid.ticks_per_beat)
                        length = core.convert.tick2loc(msg.time - note_on.time,
                                                       mid.ticks_per_beat)
                        track.add(
                            Note(
                                note_on.note,
                                velocity=note_on.velocity,
                                length=length,
                                location=location,
                                channel=note_on.channel,
                            ))
                        note_msgs[msg.channel][msg.note] = None
                else:
                    note_msgs[msg.channel][msg.note] = msg

            # These events are saved to track.events
            elif msg.type in set(('polytouch', 'control_change'
                                  'aftertouch', 'pitchwheel')):
                track.add(event)

            # If in a format 0 track, or the first track in a format 1 file, the name of the midi. Otherwise, the name of the track.
            elif msg.type == 'track_name':
                if mid.type == 0 or is_first_track:
                    mid.name = msg.name
                else:
                    track.name = msg.name

            # All copyright event save to metaevents
            elif msg.type in set(('copyright', 'lyrics')):
                mid.metaevents.add(event)

            # Track has a main `instrument`, extras will be saved to `metaevents`.
            elif msg.type == 'program_change':
                if msg.time == 0:
                    track.instrument = msg.program
                else:
                    track.add(event)

            # Midi has a main `bpm`, extras will be saved to `metaevents`.
            elif msg.type == 'set_tempo':
                if msg.time == 0:
                    mid.bpm = tempo2bpm(msg.tempo)
                else:
                    mid.metaevents.add(event)

            # Midi has a main `time_signature`, extras will be saved to `metaevents`.
            elif msg.type == 'time_signature':
                if msg.time == 0:
                    mid.time_signature = f'{msg.numerator}/{msg.denominator}'
                else:
                    mid.metaevents.add(event)

            # Midi has a main `key_signature`, extras will be saved to `metaevents`.
            elif msg.type == 'key_signature':
                if msg.time == 0:
                    mid.key_signature = msg.key
                else:
                    mid.metaevents.add(event)

        def from_type0(midi_file: MidiFile) -> 'Midi':
            mid = cls(type=0, ticks_per_beat=midi_file.ticks_per_beat)

            # Initialize tracks.
            tracks = [Track(instrument=None, channel=i) for i in range(16)]

            # Record note messages to pair.
            note_msgs = [[None] * 128 for _ in range(16)]

            # Record current time
            cur_time = 0

            for msg in midi_file.tracks[0]:
                # To absolute time.
                cur_time += msg.time
                msg.time = cur_time

                if 'channel' in msg.__dict__:
                    track = tracks[msg.channel]
                else:
                    track = None

                handel_message(msg, note_msgs, mid, track, True)

                if msg.type == 'end_of_track':
                    break

            # Drop track without notes and events
            for track in tracks:
                if len(track.notes) > 0 or len(track.events) > 0:
                    mid.tracks.append(track)

            return mid

        def from_type1(midi_file: MidiFile) -> 'Midi':
            mid = cls(type=1, ticks_per_beat=midi_file.ticks_per_beat)
            for i, mido_track in enumerate(midi_file.tracks):
                # Initialize current track
                track = Track(instrument=None, channel=None)

                # Record note messages to pair.
                note_msgs = [[None] * 128 for _ in range(16)]

                # Determine if events are all in the same channel.
                channels = set()

                # Record current time
                cur_time = 0

                for msg in mido_track:
                    # To absolute time.
                    cur_time += msg.time
                    msg.time = cur_time

                    if 'channel' in msg.__dict__:
                        channels.add(msg.channel)

                    handel_message(msg, note_msgs, mid, track, i == 0)

                    if msg.type == 'end_of_track':
                        break

                # Drop track without notes and events.
                if len(track.notes) > 0 or len(track.events) > 0:
                    mid.tracks.append(track)

                # If notes and events have a same channle, set track's channel to this one.
                if len(channels) == 1:
                    track.channel = channels.pop()

            return mid

        if midi_file.type == 0:
            return from_type0(midi_file)
        elif midi_file.type == 1:
            return from_type1(midi_file)
        elif midi_file.type == 2:
            raise MIDITypeError(f"Unsupported format 2 midi file.")

    @classmethod
    def read(cls, filename: str) -> 'Midi':
        mid = MidiFile(filename)
        mid = cls.from_mido(mid)
        return mid

    def save(self, filename) -> None:
        self.to_mido().save(filename)

    def new_track(self,
                  name: str = "",
                  channel: Optional[int] = None,
                  **kwargs) -> Track:
        # TODO. may need to be smarter..
        if not channel:
            channel = len(self.tracks)

        track = Track(name=name, channel=channel, **kwargs)
        self.tracks.append(track)
        return track



from typing import Dict, Tuple

__all__ = [
    "KEY_TO_INDEX",
    "KEYS",
    "CONTROL_DESCRIPTIONS",
    "DEGREE_TO_INDEX",
    "CONTROL_DESCRIPTIONS",
    "INSTRUMENTS",
    "DRUM_SOUNDS",
    "MESSAGE_TYPE_TO_EVENT_TYPE",
]

#  +================================================================+
#  =                             BASIC                              =
#  +================================================================+

DEFAULT_TICKS_PER_BEAT: int = 480

KEY_TO_INDEX: Dict[str, int] = {
    'C': 0,
    'C#': 1,
    'D': 2,
    'D#': 3,
    'E': 4,
    'F': 5,
    'F#': 6,
    'G': 7,
    'G#': 8,
    'A': 9,
    'A#': 10,
    'B': 11
}

KEYS: Tuple[str] = ('C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#',
                    'B')

#  +================================================================+
#  =                           FOR SCALE                            =
#  +================================================================+

# ref: https://en.wikipedia.org/wiki/List_of_musical_scales_and_modes
SCALE_NAME_TO_INTERVALS: Dict[str, Tuple[int]] = {
    '宫': (2, 2, 3, 2, 3),
    '商': (2, 3, 2, 3, 2),
    '角': (3, 2, 3, 2, 2),
    '徵': (2, 3, 2, 2, 3),
    '羽': (3, 2, 2, 3, 2),
    'Gong': (2, 2, 3, 2, 3),
    'Shang': (2, 3, 2, 3, 2),
    'Jue': (3, 2, 3, 2, 2),
    'Zhi': (2, 3, 2, 2, 3),
    'Yu': (3, 2, 2, 3, 2),
    'Ionian': (2, 2, 1, 2, 2, 2, 1),
    'Dorian': (2, 1, 2, 2, 2, 1, 2),
    'Phrygian': (1, 2, 2, 2, 1, 2, 2),
    'Lydian': (2, 2, 2, 1, 2, 2, 1),
    'Mixolydian': (2, 2, 1, 2, 2, 1, 2),
    'Aeolian': (2, 1, 2, 2, 1, 2, 2),
    'Locrian': (1, 2, 2, 1, 2, 2, 2),
}

#  +================================================================+
#  =                           FOR CHORD                            =
#  +================================================================+

CHORD_TYPE_TO_INTERVALS: Dict[str, Tuple[int]] = {
    # TODO. add more chord types
    # triad chord
    'm': (3, 4),
    'M': (4, 3),
    # seventh chord
    'm7': (3, 4, 3),
    'mM7': (3, 4, 4),
    '7': (4, 3, 3),
    'M7': (4, 3, 4),
    'm7-5': (3, 3, 4),
    # ninth chord
    'm9': (3, 4, 3, 4),
    '9': (4, 3, 3, 4),
    'M9': (4, 3, 4, 3),
    # eleventh chord
    'm11': (3, 4, 3, 4, 3),
    '11': (4, 3, 3, 4, 3),
    'M11': (4, 3, 4, 3, 3),
    # thirteenth chord
    'm13': (3, 4, 3, 4, 3, 4),
    '13': (4, 3, 3, 4, 3, 4),
    'M13': (4, 3, 4, 3, 3, 4),
    # other
    'add9': (4, 3, 7),
    'sus2': (2, 5),
    'sus4': (5, 2),
}

DEGREE_TO_INDEX: Dict[str, int] = {
    '1': 0,
    '#1': 1,
    'b2': 1,
    '2': 2,
    '#2': 3,
    'b3': 3,
    '3': 4,
    '#3': 5,
    'b4': 4,
    '4': 5,
    '#4': 6,
    'b5': 6,
    '5': 7,
    '#5': 8,
    'b6': 8,
    '6': 9,
    '#6': 10,
    'b7': 10,
    '7': 11,
    'b9': 1,
    '9': 2,
    '#9': 3,
    'b11': 4,
    '11': 5,
    '#11': 6,
    'b13': 8,
    '13': 9,
    '#13': 10,
}

#  +================================================================+
#  =                           FOR EVENT                            =
#  +================================================================+

# ref: https://www.midi.org/specifications-old/item/table-3-control-change-messages-data-bytes-2
CONTROL_DESCRIPTIONS = [
    'Bank Select',
    '* Modulation wheel',
    'Breath control',
    'Undefined',
    'Foot controller',
    'Portamento time',
    'Data Entry',
    'Channel Volume (formerly Main Volume)',
    'Balance',
    'Undefined',
    'Pan',
    'Expression Controller',
    'Effect control 1',
    'Effect control 2',
    'Undefined',
    'Undefined',
    'General Purpose Controller #1',
    'General Purpose Controller #2',
    'General Purpose Controller #3',
    'General Purpose Controller #4',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Bank Select',
    'Modulation wheel',
    'Breath control',
    'Undefined',
    'Foot controller',
    'Portamento time',
    'Data entry',
    'Channel Volume (formerly Main Volume)',
    'Balance',
    'Undefined',
    'Pan',
    'Expression Controller',
    'Effect control 1',
    'Effect control 2',
    'Undefined',
    'Undefined',
    'General Purpose Controller #1',
    'General Purpose Controller #2',
    'General Purpose Controller #3',
    'General Purpose Controller #4',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Damper pedal on/off (Sustain)',
    'Portamento on/off',
    'Sustenuto on/off',
    'Soft pedal on/off',
    'Legato Footswitch',
    'Hold 2',
    'Sound Controller 1 (Sound Variation)',
    'Sound Controller 2 (Timbre)',
    'Sound Controller 3 (Release Time)',
    'Sound Controller 4 (Attack Time)',
    'Sound Controller 5 (Brightness)',
    'Sound Controller 6',
    'Sound Controller 7',
    'Sound Controller 8',
    'Sound Controller 9',
    'Sound Controller 10',
    'General Purpose Controller #5',
    'General Purpose Controller #6',
    'General Purpose Controller #7',
    'General Purpose Controller #8',
    'Portamento Control',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Undefined',
    'Effects 1 Depth',
    'Effects 2 Depth',
    'Effects 3 Depth',
    'Effects 4 Depth',
    'Effects 5 Depth',
    'Data entry +1 ',
    'Data entry -1 ',
    'Non-Registered Parameter Number LSB',
    'Non-Registered Parameter Number MSB',
    'Registered Parameter Number LSB',
    'Registered Parameter Number MSB',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'Undefined ',
    'All Sound Off',
    'Reset All Controllers',
    'Local control on/off',
    'All notes off',
    'Omni mode off (+ all notes off)',
    'Omni mode on (+ all notes off)',
    'Poly mode on/off (+ all notes off)',
    'Poly mode on (incl mono=off +all notes off)',
]

# ref: https://www.recordingblogs.com/wiki/midi-program-change-message
PIANO: Tuple[str] = (
    'Acoustic Grand Piano',
    'Bright Acoustic Piano',
    'Electric Grand Piano',
    'Honky-tonk Piano',
    'Electric Piano 1',
    'Electric Piano 2',
    'Harpsichord',
    'Clavinet',
)

CHROMATIC_PERCUSSION: Tuple[str] = (
    'Celesta',
    'Glockenspiel',
    'Musical box',
    'Vibraphone',
    'Marimba',
    'Xylophone',
    'Tubular Bell',
    'Dulcimer',
)

ORGAN: Tuple[str] = (
    'Drawbar Organ',
    'Percussive Organ',
    'Rock Organ',
    'Church organ',
    'Reed organ',
    'Accordion',
    'Harmonica',
    'Tango Accordion',
)

GUITAR: Tuple[str] = (
    'Acoustic Guitar(nylon)',
    'Acoustic Guitar(steel)',
    'Electric Guitar(jazz)',
    'Electric Guitar(clean)',
    'Electric Guitar(muted)',
    'Overdriven Guitar',
    'Distortion Guitar',
    'Guitar harmonics',
)

BASS: Tuple[str] = (
    'Acoustic Bass',
    'Electric Bass(finger)',
    'Electric Bass(pick)',
    'Fretless Bass',
    'Slap Bass 1',
    'Slap Bass 2',
    'Synth Bass 1',
    'Synth Bass 2',
)

STRINGS: Tuple[str] = (
    'Violin',
    'Viola',
    'Cello',
    'Contrabass',
    'Tremolo Strings',
    'Pizzicato Strings',
    'Orchestral Harp',
    'Timpani',
)

ENSEMBLE: Tuple[str] = (
    'String Ensemble 1',
    'String Ensemble 2',
    'Synth Strings 1',
    'Synth Strings 2',
    'Voice Aahs',
    'Voice Oohs',
    'Synth Voice',
    'Orchestra Hit',
)

BRASS: Tuple[str] = (
    'Trumpet',
    'Trombone',
    'Tuba',
    'Muted Trumpet',
    'French horn',
    'Brass Section',
    'Synth Brass 1',
    'Synth Brass 2',
)

REED: Tuple[str] = (
    'Soprano Sax',
    'Alto Sax',
    'Tenor Sax',
    'Baritone Sax',
    'Oboe',
    'English Horn',
    'Bassoon',
    'Clarinet',
)

PIPE: Tuple[str] = (
    'Piccolo',
    'Flute',
    'Recorder',
    'Pan Flute',
    'Blown Bottle',
    'Shakuhachi',
    'Whistle',
    'Ocarina',
)

SYNTH_LEAD: Tuple[str] = (
    'Lead 1(square)',
    'Lead 2(sawtooth)',
    'Lead 3(calliope)',
    'Lead 4(chiff)',
    'Lead 5(charang)',
    'Lead 6(voice)',
    'Lead 7(fifths)',
    'Lead 8(bass + lead)',
)

SYNTH_PAD: Tuple[str] = (
    'Pad 1(new age)',
    'Pad 2(warm)',
    'Pad 3(polysynth)',
    'Pad 4(choir)',
    'Pad 5(bowed)',
    'Pad 6(metallic)',
    'Pad 7(halo)',
    'Pad 8(sweep)',
)

SYNTH_EFFECTS: Tuple[str] = (
    'FX 1(rain)',
    'FX 2(soundtrack)',
    'FX 3(crystal)',
    'FX 4(atmosphere)',
    'FX 5(brightness)',
    'FX 6(goblins)',
    'FX 7(echoes)',
    'FX 8(sci-fi)',
)

ETHNIC: Tuple[str] = (
    'Sitar',
    'Banjo',
    'Shamisen',
    'Koto',
    'Kalimba',
    'Bagpipe',
    'Fiddle',
    'Shanai',
)

PERCUSSIVE: Tuple[str] = (
    'Tinkle Bell',
    'Agogo',
    'Steel Drums',
    'Woodblock',
    'Taiko Drum',
    'Melodic Tom',
    'Synth Drum',
    'Reverse Cymbal',
)

SOUND_EFFECTS: Tuple[str] = (
    'Guitar Fret Noise',
    'Breath Noise',
    'Seashore',
    'Bird Tweet',
    'Telephone Ring',
    'Helicopter',
    'Applause',
    'Gunshot',
)

INSTRUMENTS = PIANO + CHROMATIC_PERCUSSION + ORGAN + GUITAR + \
    BASS + STRINGS + ENSEMBLE + BRASS + REED + PIPE + SYNTH_LEAD + \
    SYNTH_PAD + SYNTH_EFFECTS + ETHNIC + PERCUSSIVE + SOUND_EFFECTS

# ref: https://www.midi.org/specifications/item/gm-level-1-sound-set
DRUM_SOUNDS: Tuple[str] = (
    'Acoustic Bass Drum', 'Bass Drum 1', 'Side Stick', 'Acoustic Snare',
    'Hand Clap', 'Electric Snare', 'Low Floor Tom', 'Closed Hi Hat',
    'High Floor Tom', 'Pedal Hi Hat', 'Low Tom', 'Open Hi Hat', 'Low-Mid Tom',
    'Hi-Mid Tom', 'Crash Cymbal 1', 'High Tom', 'Ride Cymbal 1',
    'Chinese Cymbal', 'Ride Bell', 'Tambourine', 'Splash Cymbal', 'Cowbell',
    'Crash Cymbal 2', 'Vibraslap', 'Ride Cymbal 2', 'Hi Bongo', 'Low Bongo',
    'Mute Hi Conga', 'Open Hi Conga', 'Low Conga', 'High Timbale',
    'Low Timbale', 'High Agogo', 'Low Agogo', 'Cabasa', 'Maracas',
    'Short Whistle', 'Long Whistle', 'Short Guiro', 'Long Guiro', 'Claves',
    'Hi Wood Block', 'Low Wood Block', 'Mute Cuica', 'Open Cuica',
    'Mute Triangle', 'Open Triangle')

MESSAGE_TYPE_TO_EVENT_TYPE: Dict[str, str] = {
    'note_off': 'NoteOff',
    'note_on': 'NoteOn',
    'polytouch': 'NotePressure',
    'control_change': 'ControlChange',
    'program_change': 'ProgramChange',
    'aftertouch': 'ChannelPressure',
    'pitchwheel': 'PitchBend',
    'text': 'Text',
    'track_name': 'TrackName',
    'lyrics': 'Lyric',
    'marker': 'Marker',
    'cue_marker': 'CuePoint',
    'set_tempo': 'SetBpm',
    'time_signature': 'TimeSignature',
    'key_signature': 'KeySignature',
}

EVENT_TYPE_TO_MESSAGE_TYPE: Dict[str, str] = dict(
    zip(MESSAGE_TYPE_TO_EVENT_TYPE.values(),
        MESSAGE_TYPE_TO_EVENT_TYPE.keys()))



from itertools import accumulate
from typing import Sequence, Union

from ..constants import DEGREE_TO_INDEX, SCALE_NAME_TO_INTERVALS
from ..core.clip import Clip
from ..core.convert import key2index
from ..exceptions import ParameterError
from ..warps import optional_inplace

__all__ = ["scale_map"]


@optional_inplace(False)
def scale_map(clip: Clip,
              key: str = 'C',
              scale: Union[str, Sequence[Union[int, str]]] = "") -> Clip:
    if isinstance(scale, str):
        itvs = SCALE_NAME_TO_INTERVALS[scale]
        idxs = [0]
        for itv in itvs:
            idxs.append(idxs[-1] + itv)
        mapping = []
        for i in range(len(idxs) - 1):
            cur = idxs[i]
            nxt = idxs[i + 1]
            mapping.append(cur)
            r = (nxt - cur - 1) // 2
            l = (nxt - cur - 1) - r
            mapping += [cur] * l + [nxt] * r

    elif isinstance(scale, Sequence):
        if isinstance(scale[0], int):
            mapping = scale
        elif isinstance(scale[0], str):
            mapping = [DEGREE_TO_INDEX[degree] for degree in scale]
        else:
            ParameterError(f"Invalid value of scale: {scale}")
    else:
        ParameterError(f"Invalid value of scale: {scale}")

    for note in clip.notes:
        idx = key2index(note.key, key)
        note += mapping[idx] - idx
    return clip



import math
from ast import Slice
from functools import partialmethod
from itertools import chain
from typing import Iterable, List, Optional, Union

from mido import Message

from ..constants import DEFAULT_TICKS_PER_BEAT
from ..exceptions import ParameterError
from ..warps import optional_inplace
from .event import ChannelEvent, Event, EventSet
from .note import Note, NoteSet

__all__ = [
    "Clip",
]


class Clip:

    def __init__(self,
                 items: Optional[Iterable[Union[Note, Event, 'Clip']]] = None,
                 **kwargs) -> None:
        self.notes = NoteSet()
        self.events = EventSet()
        if items:
            for item in items:
                self.add(item)

    def __iter__(self):
        for item in chain(self.notes, self.events):
            yield item

    @property
    def n_notes(self):
        return len(self.notes)

    @property
    def n_events(self):
        return len(self.events)

    @property
    def notes_length(self) -> float:
        if self.notes:
            last_note = self.notes[-1]
            return math.ceil(last_note.location + last_note.length)
        else:
            return 0

    @property
    def events_length(self) -> float:
        if self.events:
            last_event = self.events[-1]
            return math.ceil(last_event.location)
        else:
            return 0

    @property
    def length(self) -> float:
        return max(self.notes_length, self.events_length)


    def __setitem__(self, key: Slice, item: Union[Event, Note,
                                                  'Clip']) -> None:
        if isinstance(key, slice):
            lloc = key.start if key.start else 0
            rloc = key.stop if key.stop else self.length
            self.clear(lloc, rloc)
            self.add(item, lloc)
        else:
            raise Exception(f"Invalid key type: {type(key)}.")

    def __getitem__(self, key: Slice) -> 'Clip':
        if isinstance(key, slice):
            return self.slice(key.start, key.stop, inplace=False)
        else:
            raise ParameterError(f"Invalid key type : {type(key)}.")

    def __contains__(self, item: Union[Note, Event]) -> bool:
        if isinstance(item, Note):
            return item in self.notes
        if isinstance(item, Event):
            return item in self.events

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(n_notes={self.n_notes}, n_events={self.n_events}, length={self.length})'

    def copy_notes(self):
        return NoteSet(note.copy() for note in self.notes)

    def copy_events(self):
        return EventSet(event.copy() for event in self.events)

    def copy(self):
        clip = self.__class__()
        clip.notes = self.copy_notes()
        clip.events = self.copy_events()
        return clip

    @optional_inplace(True)
    def slice(self,
              lloc: Optional[float] = None,
              rloc: Optional[float] = None) -> 'Clip':
        lloc = 0 if lloc is None else lloc
        rloc = self.length if rloc is None else rloc
        reverse = False
        if lloc > rloc:
            lloc, rloc = rloc, lloc
            reverse = True

        l = self.notes.bisect_left(Note(0, location=lloc))
        r = self.notes.bisect_left(Note(0, location=rloc))
        notes = self.notes[l:r]
        for note in notes:
            note <<= lloc
        self.notes = NoteSet(notes)

        l = self.events.bisect_left(ChannelEvent(location=lloc))
        r = self.events.bisect_left(ChannelEvent(location=rloc))
        events = self.events[l:r]
        for event in events:
            event <<= lloc
        self.events = EventSet(events)
        if reverse:
            self.reverse()
        return self

    def _add_event(self,
                   event: Event,
                   location: Optional[float] = None) -> None:
        if location:
            event.location = location
        self.events.add(event)

    def _add_note(self, note: Note, location: Optional[float] = None) -> None:
        if location:
            note.location = location
        self.notes.add(note)

    def _add_clip(self,
                  clip: 'Clip',
                  location: Optional[float] = None) -> None:
        for item in clip:
            if location:
                item.location += location
            if isinstance(item, Event):
                self._add_event(item)
            elif isinstance(item, Note):
                self._add_note(item)

    def up(self, n: int) -> 'Clip':
        for note in self.notes:
            note += n
        return self

    trans = up

    @optional_inplace(True)
    def add(
        self,
        item: Union[int, Event, Note, 'Clip'],
        location: Optional[float] = None,
    ) -> 'Clip':

        if 'copy' in item.__dir__():
            item = item.copy()
        if isinstance(item, Event):
            self._add_event(item, location)
        elif isinstance(item, Note):
            self._add_note(item, location)
        elif isinstance(item, Clip):
            self._add_clip(item, location)
        elif isinstance(item, int):
            self.up(item)
        return self

    __add__ = partialmethod(add, location=None, inplace=False)
    __iadd__ = partialmethod(add, location=None, inplace=True)

    @optional_inplace(True)
    def down(self, n: int) -> 'Clip':
        for note in self.notes:
            note -= n
        return self

    @optional_inplace(True)
    def sub(
        self,
        item: Union[int, Event, Note],
    ) -> 'Clip':

        if 'copy' in item.__dir__():
            item = item.copy()
        if isinstance(item, Event):
            self.events.remove(item)
        elif isinstance(item, Note):
            self.notes.remove(item)
        elif isinstance(item, int):
            self.down(item)
        return self

    __sub__ = partialmethod(sub, inplace=False)
    __isub__ = partialmethod(sub, inplace=True)

    @optional_inplace(True)
    def remove(
        self,
        item: Union[Event, Note],
    ) -> 'Clip':
        if isinstance(item, Event):
            self.events.remove(item)
        elif isinstance(item, Note):
            self.notes.remove(item)
        return self

    @optional_inplace(True)
    def discard(
        self,
        item: Union[Event, Note],
    ) -> 'Clip':
        if isinstance(item, Event):
            self.events.discard(item)
        elif isinstance(item, Note):
            self.notes.discard(item)
        return self

    @optional_inplace(True)
    def concat(self, clip: 'Clip') -> 'Clip':
        self._add_clip(clip.copy(), self.length)
        return self

    __or__ = partialmethod(concat, inplace=False)
    __ior__ = partialmethod(concat, inplace=True)

    @optional_inplace(True)
    def extend(self, clips: Iterable['Clip']) -> 'Clip':
        for clip in clips:
            self.concat(clip)
        return self

    @optional_inplace(True)
    def repeat(self, n: int) -> 'Clip':
        clip = self.copy()
        for _ in range(n - 1):
            self |= clip
        return self

    __pow__ = partialmethod(repeat, inplace=False)
    __ipow__ = partialmethod(repeat, inplace=True)

    def shift(self, length: float) -> 'Clip':
        for item in self:
            item >>= length
        return self

    @optional_inplace(True)
    def lshift(self, length: float) -> 'Clip':
        self.shift(-length)
        return self

    __lshift__ = partialmethod(lshift, inplace=False)
    __ilshift__ = partialmethod(lshift, inplace=True)

    @optional_inplace(True)
    def rshift(self, length: float) -> 'Clip':
        self.shift(length)
        return self

    __rshift__ = partialmethod(rshift, inplace=False)
    __irshift__ = partialmethod(rshift, inplace=True)

    @optional_inplace(True)
    def at(self, channel: int) -> 'Clip':
        for item in self:
            if 'channel' in item.__dir__():
                item.channel = channel
        return self

    __matmul__ = partialmethod(at, inplace=False)
    __imatmul__ = partialmethod(at, inplace=True)

    @optional_inplace(True)
    def reverse(self) -> 'Clip':
        length = self.length
        for item in self:
            if isinstance(item, Note):
                item.location = length - (item.location + item.length)
            else:
                item.location = length - item.location
        return self

    __reversed__ = partialmethod(reverse, inplace=False)

    @optional_inplace(True)
    def invert(self) -> 'Clip':
        max_value = max(note.value for note in self.notes)
        min_value = min(note.value for note in self.notes)
        for note in self.notes:
            note.value = min_value + max_value - note.value
        return self

    __invert__ = partialmethod(invert, inplace=False)

    @optional_inplace(True)
    def zoom(self, ration: float) -> 'Clip':
        for item in self:
            item ^= ration
        return self

    __xor__ = partialmethod(zoom, inplace=False)
    __ixor__ = partialmethod(zoom, inplace=True)

    @optional_inplace(True)
    def scale(self, ration: float) -> 'Clip':
        for note in self.notes:
            note *= ration
        return self

    __mul__ = partialmethod(scale, inplace=False)
    __imul__ = partialmethod(scale, inplace=True)

    @optional_inplace(True)
    def decay(self, ration: float) -> 'Clip':
        for note in self.notes:
            note &= ration
        return self

    __and__ = partialmethod(decay, inplace=False)
    __iand__ = partialmethod(decay, inplace=True)

    @optional_inplace(True)
    def quantify(self, precision: int) -> 'Clip':
        for item in self:
            item.location = math.ceil(item.location / precision) * precision
        return self

    def clear(self,
              lloc: Optional[float] = None,
              rloc: Optional[float] = None) -> 'Clip':
        lloc = 0 if lloc is None else lloc
        rloc = self.length if rloc is None else rloc
        notes = NoteSet()
        events = EventSet()
        for item in self:
            loc = item.location
            if not (loc >= lloc and loc < rloc):
                notes.add(item)
                events.add(item)
        self.notes = notes
        self.events = events
        return self

    def _to_reltime(self, msgs: List[Message]) -> None:
        now = 0
        for i in range(len(msgs)):
            msgs[i].time -= now
            now += msgs[i].time

    @property
    def msgs(self) -> List[Message]:
        return self.messages()

    def messages(self,
                 ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT,
                 mode='reltime') -> List[Message]:
        msgs = []
        for note in self.notes:
            if note.location >= 0:
                msgs.extend(note.messages(ticks_per_beat))
        for event in self.events:
            if event.location >= 0:
                msgs.append(event.message(ticks_per_beat))

        msgs.sort(key=lambda msg: (msg.time, msg.__dict__.setdefault(
            "note", 0), msg.__dict__.setdefault("velocity", 0)))
        if mode == 'reltime':
            self._to_reltime(msgs)
        return msgs



from .core import *
from .devices import *
from .effects import *
from .generator import *
from .utils import *
from .constants import *

__all__ = []
__all__.extend(core.__all__)
__all__.extend(devices.__all__)
__all__.extend(effects.__all__)
__all__.extend(generator.__all__)
__all__.extend(utils.__all__)
__all__.extend(constants.__all__)


__title__ = 'infmidi'
__version__ = '0.1.2'
__author__ = 'gongyibei'



import random
from ..core.clip import Clip

__all__ = []

def random_velocity(clip:Clip, min_velocity: int=0, max_velocity: int=127) -> Clip:
    for note in clip.notes:
        note.velocity = random.randint(min_velocity, max_velocity)
    return clip

def uniform_velocity(clip:Clip, min_velocity: int=0, max_velocity: int=127) -> Clip:
    for note in clip.notes:
        note.velocity = random.uniform(min_velocity, max_velocity)
    return clip


import operator
from functools import partialmethod, total_ordering
from typing import Iterable, Optional, Union

from mido import Message, MetaMessage, bpm2tempo, tempo2bpm
from sortedcontainers import SortedSet

from .. import core
from ..constants import (CONTROL_DESCRIPTIONS, DEFAULT_TICKS_PER_BEAT,
                         INSTRUMENTS, KEYS)
from ..warps import optional_inplace

BASIC = ["Event"]

CHANNEL_EVENT = [
    "ChannelEvent", "NoteOff", "NoteOn", "NotePressure", "ControlChange",
    "ProgramChange", "ChannelPressure", "PitchBend"
]
META_EVENT = [
    "MetaEvent", "Text", "Copyright", "Lyric", "Marker", "CuePoint", "TrackName", "SetBpm",
    "TimeSignature", "KeySignature", "SequencerSpecific"
]
EXTRA = ["EventSet"]

__all__ = BASIC + CHANNEL_EVENT + META_EVENT + EXTRA


@total_ordering
class Event:

    def __init__(self, location: float = 0) -> None:
        self.location = location

    def _compare(self, a, b, op):
        _a = a.location
        _b = b.location
        return op(_a, _b)

    def __lt__(self, other: 'Event') -> bool:
        return self._compare(self, other, operator.lt)

    def __eq__(self, other: 'Event') -> bool:
        return self._compare(self, other, operator.eq)

    def shift(self, length: float = 0) -> 'Event':
        self.location += length
        return self

    @optional_inplace(True)
    def lshift(self, length: float = 0) -> 'Event':
        return self.shift(-length)

    __lshift__ = partialmethod(lshift, inplace=False)
    __ilshift__ = partialmethod(lshift, inplace=True)

    @optional_inplace(True)
    def rshift(self, length: float = 0) -> 'Event':
        return self.shift(length)

    __rshift__ = partialmethod(rshift, inplace=False)
    __irshift__ = partialmethod(rshift, inplace=True)

    @optional_inplace(True)
    def zoom(self, ration: float = 1) -> 'Event':
        self.location *= ration
        return self

    __xor__ = partialmethod(zoom, inplace=False)
    __ixor__ = partialmethod(zoom, inplace=True)

    @property
    def msg(self):
        return self.message()

    def copy(self):
        raise NotImplementedError

    def message(self):
        raise NotImplementedError

    def from_message(self):
        raise NotImplementedError


class ChannelEvent(Event):

    def __init__(self, location: float = 0, channel: int = 0) -> None:
        super().__init__(location)

        self.channel = channel

    def _compare(self, a, b, op):
        _a = (a.location, a.channel)
        _b = (b.location, b.channel)
        return op(_a, _b)

    def __hash__(self):
        return hash((self.__class__, self.location, self.channel))

    @optional_inplace(True)
    def at(self, channel: int) -> 'ChannelEvent':
        self.channel = channel
        return self

    __matmul__ = partialmethod(at, inplace=False)
    __imatmul__ = partialmethod(at, inplace=True)


class _NoteEvent(ChannelEvent):

    def __init__(
        self,
        note: Union[str, int, float],
        location: float = 0.,
        channel: int = 0,
    ) -> None:
        super().__init__(location, channel)
        if isinstance(note, str):
            self.value = core.convert.name2value(note)
        elif isinstance(note, int):
            self.value = note
        elif isinstance(note, float):
            self.value = core.convert.frequency2value(note)
        else:
            raise Exception("Unsupported note type！")
        self.channel = channel

    @optional_inplace(True)
    def up(self, n: int) -> '_NoteEvent':
        self.value += n
        return self

    __add__ = partialmethod(up, inplace=False)
    __iadd__ = partialmethod(up, inplace=True)

    @optional_inplace(True)
    def down(self, n: int) -> '_NoteEvent':
        self.value -= n
        return self

    __sub__ = partialmethod(down, inplace=False)
    __isub__ = partialmethod(down, inplace=True)


    @property
    def name(self) -> str:
        return core.convert.value2name(self.value)

    @name.setter
    def name(self, other: str) -> None:
        self.value = core.convert.name2value(other)

    @property
    def frequency(self) -> float:
        return core.convert.value2frequency(self.value)

    @frequency.setter
    def frequency(self, other: float) -> None:
        self.value = core.convert.frequency2value(other)
    
    @property
    def key(self) -> float:
        return KEYS[self.value%12]
    
    @property
    def octave(self) -> float:
        return self.value // 12 - 1

    def __hash__(self):
        return hash((self.__class__, self.value, self.location, self.channel))


class NoteOff(_NoteEvent):
    message_type = "note_off"

    def __init__(
        self,
        note: Union[str, int, float],
        velocity: int = 127,
        location: float = 0.,
        channel: int = 0,
    ) -> None:
        super().__init__(note, location=location, channel=channel)
        self.velocity = velocity

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name="{self.name}", value={self.value}, frequency={self.frequency:.2f}, velocity={self.velocity}, location={self.location:.2f}, channel={self.channel})'

    def message(self, ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> Message:
        return Message('note_off',
                       channel=self.channel,
                       note=self.value,
                       velocity=self.velocity,
                       time=core.convert.loc2tick(self.location,
                                                  ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: Message,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'NoteOff':
        return cls(msg.note,
                   velocity=msg.velocity,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat),
                   channel=msg.channel)

    def _compare(self, a, b, op):
        _a = (a.location, a.value, a.velocity, a.channel)
        _b = (b.location, b.value, b.velocity, b.channel)
        return op(_a, _b)

    def copy(self) -> 'NoteOff':
        return self.__class__(self.value,
                              velocity=self.velocity,
                              location=self.location,
                              channel=self.channel)


class NoteOn(_NoteEvent):
    message_type = "note_on"

    def __init__(
        self,
        note: Union[str, int, float],
        velocity: int = 127,
        location: float = 0.,
        channel: int = 0,
    ) -> None:
        super().__init__(note, location=location, channel=channel)
        self.velocity = velocity

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name="{self.name}", value={self.value}, frequency={self.frequency:.2f}, velocity={self.velocity}, location={self.location:.2f}, channel={self.channel})'

    def message(self, ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> Message:
        return Message('note_on',
                       channel=self.channel,
                       note=self.value,
                       velocity=self.velocity,
                       time=core.convert.loc2tick(self.location,
                                                  ticks_per_beat))

    @classmethod
    def from_message(cls,
                     msg: Message,
                     ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'NoteOn':
        return cls(msg.note,
                   velocity=msg.velocity,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat),
                   channel=msg.channel)

    def _compare(self, a, b, op):
        _a = (a.location, a.value, a.velocity, a.channel)
        _b = (b.location, b.value, b.velocity, b.channel)
        return op(_a, _b)

    def copy(self) -> 'NoteOn':
        return self.__class__(self.value,
                              velocity=self.velocity,
                              location=self.location,
                              channel=self.channel)


class NotePressure(_NoteEvent):
    message_type = "polytouch"

    def __init__(
        self,
        note: Union[str, int, float],
        pressure: int = 127,
        location: float = 0.,
        channel: int = 0,
    ) -> None:
        super().__init__(note, location=location, channel=channel)
        self.pressure = pressure

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name="{self.name}", value={self.value}, frequency={self.frequency:.2f}, pressure={self.pressure}, location={self.location:.2f}, channel={self.channel})'

    def message(self, ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> Message:
        return Message('polytouch',
                       channel=self.channel,
                       note=self.value,
                       value=self.pressure,
                       time=core.convert.loc2tick(self.location,
                                                  ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: Message,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'NotePressure':
        return cls(msg.note,
                   pressure=msg.pressure,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat),
                   channel=msg.channel)

    def _compare(self, a, b, op):
        _a = (a.location, a.value, a.pressure, a.channel)
        _b = (b.location, b.value, b.pressure, b.channel)
        return op(_a, _b)

    def copy(self) -> 'NotePressure':
        return self.__class__(self.value,
                              pressure=self.pressure,
                              location=self.location,
                              channel=self.channel)


class ControlChange(ChannelEvent):
    message_type = "control_change"

    def __init__(self,
                 control: int,
                 value: int = 0,
                 location: float = 0,
                 channel: int = 0) -> None:
        super().__init__(location, channel)
        self.control = control
        self.value = value

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(control={self.control}, value={self.value}, description="{self.description}", location={self.location:.2f}, channel={self.channel})'

    @property
    def description(self):
        return CONTROL_DESCRIPTIONS[self.control]

    def message(self, ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> Message:
        return Message('control_change',
                       channel=self.channel,
                       control=self.control,
                       value=self.value,
                       time=core.convert.loc2tick(self.location,
                                                  ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: Message,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'ControlChange':
        return cls(msg.control,
                   msg.value,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat),
                   channel=msg.channel)

    def copy(self) -> 'ControlChange':
        return self.__class__(self.control,
                              self.value,
                              location=self.location,
                              channel=self.channel)


class ProgramChange(ChannelEvent):
    message_type = "program_change"

    def __init__(self,
                 inst: Union[str, int] = 0,
                 location: float = 0,
                 channel: int = 0) -> None:
        super().__init__(location, channel)
        if isinstance(inst, int):
            self.id = inst
        elif isinstance(inst, str):
            self.id = self._fuzzy_match(inst)

    @property
    def name(self) -> str:
        return INSTRUMENTS[self.id]

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'{self.__class__.__name__}(id={self.id}, name="{self.name}", location={self.location:.2f}, channel={self.channel})'

    def _fuzzy_match(self, inst: str):
        # TODO.
        for i, name in enumerate(INSTRUMENTS):
            if name == inst:
                return i
        return 0

    def message(self, ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> Message:
        return Message('program_change',
                       channel=self.channel,
                       program=self.id,
                       time=core.convert.loc2tick(self.location,
                                                  ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: Message,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'ProgramChange':
        return cls(msg.program,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat),
                   channel=msg.channel)

    def copy(self) -> 'ProgramChange':
        return self.__class__(self.id,
                              location=self.location,
                              channel=self.channel)

    @staticmethod
    def print_instruments():
        for idx, instrument in enumerate(INSTRUMENTS):
            print(f'{idx:<3d} {instrument}')


class ChannelPressure(ChannelEvent):
    message_type = "aftertouch"

    def __init__(self,
                 pressure: float = 0,
                 location: float = 0,
                 channel: int = 0) -> None:
        super().__init__(location, channel)
        self.pressure = pressure

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(pressure={self.pressure}, location={self.location:.2f}, channel={self.channel})'

    def message(self, ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> Message:
        return Message('aftertouch',
                       channel=self.channel,
                       value=self.pressure,
                       time=core.convert.loc2tick(self.location,
                                                  ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: Message,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'ChannelPressure':
        return cls(msg.value,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat),
                   channel=msg.channel)

    def copy(self) -> 'ChannelPressure':
        return self.__class__(self.pressure,
                              location=self.location,
                              channel=self.channel)


class PitchBend(ChannelEvent):
    message_type = "pitchwheel"

    def __init__(self,
                 pitch: float = 0.,
                 location: float = 0.,
                 channel: int = 0) -> None:
        super().__init__(location, channel)
        self.pitch = max(min(pitch, 1.), -1.)

    def __str__(self):
        return self.pitch

    def __repr__(self):
        return f'{self.__class__.__name__}(pitch={self.pitch:.2f}, location={self.location:.2f}, channel={self.channel})'

    def message(self, ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> Message:
        return Message('pitchwheel',
                       pitch=int(self.pitch * 8191),
                       channel=self.channel,
                       time=core.convert.loc2tick(self.location,
                                                  ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: Message,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'PitchBend':
        if msg.pitch < 0:
            pitch = msg.pitch / 8192
        else:
            pitch = msg.pitch / 8191
        return cls(pitch,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat),
                   channel=msg.channel)

    def copy(self) -> 'PitchBend':
        return self.__class__(self.pressure,
                              location=self.location,
                              channel=self.channel)

    def copy(self) -> 'PitchBend':
        return self.__class__(self.pitch,
                              location=self.location,
                              channel=self.channel)


class MetaEvent(Event):
    pass


class _TextMetaEvent(MetaEvent):

    def __init__(self, text="", location: float = 0.):
        super().__init__(location)
        self.text = text

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(text="{repr(self.text)[1:-1]}", location={self.location:.2f})'

    def message(self, ticks_per_beat=DEFAULT_TICKS_PER_BEAT) -> MetaMessage:
        return MetaMessage(self.message_type,
                           text=self.text,
                           time=core.convert.loc2tick(self.location,
                                                      ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: MetaMessage,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> '_TextMetaEvent':
        return cls(msg.text,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat))

    def copy(self) -> '_TextMetaEvent':
        return self.__class__(self.text, location=self.location)

    def __hash__(self):
        return hash((self.__class__, self.text, self.location))


class Text(_TextMetaEvent):
    message_type = "text"


class Copyright(_TextMetaEvent):
    message_type = "copyright"


class Lyric(_TextMetaEvent):
    message_type = "lyrics"


class Marker(_TextMetaEvent):
    message_type = "marker"


class CuePoint(_TextMetaEvent):
    message_type = "cue_marker"


class TrackName(MetaEvent):
    message_type = "track_name"

    def __init__(self, name="", location: float = 0.):
        super().__init__(location)
        self.name = name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name="{repr(self.name)[1:-1]}", location={self.location:.2f})'

    def message(self, ticks_per_beat=DEFAULT_TICKS_PER_BEAT) -> MetaMessage:
        return MetaMessage(self.message_type,
                           name=self.name,
                           time=core.convert.loc2tick(self.location,
                                                      ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: MetaMessage,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'TrackName':
        return cls(msg.name,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat))

    def copy(self) -> 'TrackName':
        return self.__class__(self.name, location=self.location)

    def __hash__(self):
        return hash((self.__class__, self.name, self.location))


class SetBpm(MetaEvent):
    message_type = "set_tempo"

    def __init__(self, bpm: float = 120., location: float = 0.):
        super().__init__(location)
        self.bpm = bpm

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(bpm={self.bpm:.2f}, location={self.location:.2f})'

    def message(self, ticks_per_beat=DEFAULT_TICKS_PER_BEAT) -> MetaMessage:
        return MetaMessage("set_tempo",
                           tempo=bpm2tempo(self.bpm),
                           time=core.convert.loc2tick(self.location,
                                                      ticks_per_beat))

    @classmethod
    def from_message(cls,
                     msg: MetaMessage,
                     ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'SetBpm':
        return cls(bpm=tempo2bpm(msg.tempo),
                   location=core.convert.tick2loc(msg.time, ticks_per_beat))

    def copy(self) -> 'SetBpm':
        return self.__class__(self.bpm, location=self.location)

    def __hash__(self):
        return hash((self.__class__, self.bpm, self.location))


class TimeSignature(MetaEvent):
    message_type = "time_signature"

    def __init__(self, signature: str = "4/4", location: float = 0.):
        super().__init__(location)
        self.signature = signature

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(signature="{self.signature}", location={self.location:.2f})'

    @property
    def numerator(self) -> int:
        return int(self.signature.split('/')[0])

    @property
    def denominator(self) -> int:
        return int(self.signature.split('/')[1])

    def message(self, ticks_per_beat=DEFAULT_TICKS_PER_BEAT) -> MetaMessage:
        return MetaMessage("time_signature",
                           numerator=self.numerator,
                           denominator=self.denominator,
                           time=core.convert.loc2tick(self.location,
                                                      ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: MetaMessage,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'TimeSignature':
        return cls(signature=f"{msg.numerator}/{msg.denominator}",
                   location=core.convert.tick2loc(msg.time, ticks_per_beat))

    def copy(self) -> 'TimeSignature':
        return self.__class__(self.signature, location=self.location)

    def __hash__(self):
        return hash((self.__class__, self.signature, self.location))


class KeySignature(MetaEvent):
    message_type = "key_signature"

    def __init__(self, signature: str = 'C', location: float = 0.):
        super().__init__(location)
        self.signature = signature

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(signature="{self.signature}", location={self.location:.2f})'

    def message(self, ticks_per_beat=DEFAULT_TICKS_PER_BEAT) -> MetaMessage:
        return MetaMessage("key_signature",
                           key=self.signature,
                           time=core.convert.loc2tick(self.location,
                                                      ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: MetaMessage,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'KeySignature':
        return cls(signature=msg.key,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat))

    def copy(self) -> 'KeySignature':
        return self.__class__(self.signature, location=self.location)

    def __hash__(self):
        return hash((self.__class__, self.signature, self.location))


class SequencerSpecific(MetaEvent):
    message_type = "sequencer_specific"

    def __init__(self, data, location: float = 0) -> None:
        super().__init__(location)
        self.data = data

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(data={self.data}, location={self.location:.2f})'
    
    def message(self, ticks_per_beat=DEFAULT_TICKS_PER_BEAT) -> MetaMessage:
        return MetaMessage("sequencer_specific",
                           key=self.signature,
                           time=core.convert.loc2tick(self.location,
                                                      ticks_per_beat))

    @classmethod
    def from_message(
            cls,
            msg: MetaMessage,
            ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> 'KeySignature':
        return cls(data=msg.data,
                   location=core.convert.tick2loc(msg.time, ticks_per_beat))
   
    def copy(self) -> 'SequencerSpecific':
        return self.__class__(self.data, location=self.location) 
    
    def __hash__(self):
        return hash((self.__class__, self.location))
    


class EventSet(SortedSet):

    def __init__(self, iterable: Optional[Iterable[Event]] = None, key=None):
        super().__init__(iterable=iterable, key=key)

    # borrowed from mido
    def __repr__(self):
        if len(self) == 0:
            events = ''
        elif len(self) == 1:
            events = f'[{repr(self[0])}]'
        else:
            events = f',\n  '.join(repr(m) for m in self)
            events = f'[\n  {events}\n]'
        return f'{self.__class__.__name__}({events})'

    __str__ = __repr__

    def __iter__(self):
        for event in self._list:
            yield event
        # the yielded note can be modifield, to update the order of events in SortedSet
        self.fix()

    def fix(self) -> None:
        events = list(self._list)
        self.clear()
        self.update(events)

    def add(self, value: Event) -> None:
        if isinstance(value, Event):
            if value in self:
                self.discard(value)
            super().add(value)

    _add = add



from typing import Optional, Sequence, Union

from .. import Clip, Note
from ..constants import CHORD_TYPE_TO_INTERVALS, DEGREE_TO_INDEX
from ..exceptions import ParameterError

__all__ = [
    "chord",
]


def _chord_from_type(root: str,
                     chord_type: str,
                     gap: float = 0,
                     length: float = 1.) -> Clip:
    intervals = CHORD_TYPE_TO_INTERVALS[chord_type]
    return _chord_from_intervals(root, intervals, gap, length)


def _chord_from_intervals(root: str,
                          intervals: Sequence[int],
                          gap: float = 0.,
                          length: float = 1.) -> Clip:
    clip = Clip()
    note = Note(root, length=length)
    clip.add(note)
    loc = 0
    for itv in intervals:
        note = note + itv
        loc += gap
        clip.add(note, loc)

    return clip


def _chord_from_degrees(root: str,
                        degrees: Sequence[str],
                        gap: float = 0.,
                        length: float = 1.) -> Clip:
    indexes = [DEGREE_TO_INDEX[num] for num in degrees]
    intervals = []
    for i in range(len(indexes) - 1):
        itv = indexes[i + 1] - indexes[i]
        itv = (itv + 11) % 12 + 1
        intervals.append(itv)
    root = (Note(root) + indexes[0]).name
    return _chord_from_intervals(root, intervals, gap, length)


def chord(*args, gap: float = 0., length: float = 1.) -> Optional[Clip]:
    if len(args) == 1 and isinstance(args[0], str):
        root, chd = args[0].split(':')
        return _chord_from_type(root, chd, gap, length)
    elif len(args) == 2:
        root, chd = args
        if isinstance(chd, str):
            return _chord_from_type(root, chd, gap, length)
        elif isinstance(chd, Sequence):
            if isinstance(chd[0], int):
                return _chord_from_intervals(root, chd, gap, length)
            elif isinstance(chd[0], str):
                return _chord_from_degrees(root, chd, gap, length)
    else:
        raise ParameterError("The lenght of args shoud be less than 2.")



from ..core import Clip

import http.server
from threading import Thread
from functools import partial
import asyncio
import websockets
import json
import threading
import pkg_resources
from typing import Optional

__all__ = ['Viewer']


def ServeDirectoryWithHTTP(port, directory="./INFMIDIViewer/"):
    if port:
        addr = ("localhost", port)
    else:
        addr = ("localhost", 0)

    directory = pkg_resources.resource_filename( __name__, 'INFMIDIViewer')
    handler = partial(http.server.SimpleHTTPRequestHandler, directory=directory)
    httpd = http.server.HTTPServer(addr, handler, False)
    httpd.allow_reuse_address = True
    httpd.timeout = 0.5
    httpd.server_bind()
    httpd.server_activate()

    print(f"Open this link in your browser: http://{httpd.server_name}:{httpd.server_port}")

    def serve_forever(httpd):
        with httpd:  # to make sure httpd.server_close is called
            httpd.serve_forever()

    thread = Thread(target=serve_forever, args=(httpd, ))
    thread.setDaemon(True)
    thread.start()

class Viewer:
    def __init__(self, clip: Clip, port:Optional[int]=None):
        self.clip = clip
        ServeDirectoryWithHTTP(port)
        threading.Thread(target=self.run).start()
    
    async def communicate(self, websocket):
        while True:
            notes = []
            for note in self.clip.notes:
                notes.append([note.value, note.velocity, note.location, note.length])
            event = {"notes": notes}
            await asyncio.sleep(0.1)
            await websocket.send(json.dumps(event))


    async def main(self):
        async with websockets.serve(self.communicate, "localhost", 8765):
            await asyncio.Future()

    def run(self):
        asyncio.run(self.main())
